@@ -33,21 +33,63 @@
 
 #include "gdm-common.h"
 
-void
-gdm_set_fatal_warnings_if_unstable (void)
+gboolean
+gdm_is_version_unstable (void)
 {
-        char **versions;
+        char   **versions;
+        gboolean unstable;
+
+        unstable = FALSE;
 
         versions = g_strsplit (VERSION, ".", 3);
         if (versions && versions [0] && versions [1]) {
                 int major;
                 major = atoi (versions [1]);
                 if ((major % 2) != 0) {
-                        g_setenv ("G_DEBUG", "fatal_criticals", FALSE);
-                        g_log_set_always_fatal (G_LOG_LEVEL_CRITICAL);
+                        unstable = TRUE;
                 }
         }
         g_strfreev (versions);
+
+        return unstable;
+}
+
+void
+gdm_set_fatal_warnings_if_unstable (void)
+{
+        if (gdm_is_version_unstable ()) {
+                g_setenv ("G_DEBUG", "fatal_criticals", FALSE);
+                g_log_set_always_fatal (G_LOG_LEVEL_CRITICAL);
+        }
+}
+
+int
+gdm_wait_on_pid (int pid)
+{
+        int status;
+
+ wait_again:
+        errno = 0;
+        if (waitpid (pid, &status, 0) < 0) {
+                if (errno == EINTR) {
+                        goto wait_again;
+                } else if (errno == ECHILD) {
+                        ; /* do nothing, child already reaped */
+                } else {
+                        g_debug ("GdmCommon: waitpid () should not fail");
+                }
+        }
+
+        g_debug ("GdmCommon: process (pid:%d) done (%s:%d)",
+                 (int) pid,
+                 WIFEXITED (status) ? "status"
+                 : WIFSIGNALED (status) ? "signal"
+                 : "unknown",
+                 WIFEXITED (status) ? WEXITSTATUS (status)
+                 : WIFSIGNALED (status) ? WTERMSIG (status)
+                 : -1);
+
+        return status;
 }
 
 int
@@ -57,7 +99,8 @@ gdm_signal_pid (int pid,
         int status = -1;
 
         /* perhaps block sigchld */
-        g_debug ("sending signal %d to process %d", signal, pid);
+        g_debug ("GdmCommon: sending signal %d to process %d", signal, pid);
+        errno = 0;
         status = kill (pid, signal);
 
         if (status < 0) {
@@ -311,6 +354,7 @@ _read_bytes (int      fd,
         do {
                 size_t bytes_read = 0;
 
+                errno = 0;
                 bytes_read = read (fd, ((guchar *) bytes) + total_bytes_read,
                                    bytes_left_to_read);
 
@@ -364,6 +408,7 @@ gdm_generate_random_bytes (gsize    size,
          * than the passed in size.
          */
 
+        errno = 0;
         fd = open ("/dev/urandom", O_RDONLY);
 
         if (fd < 0) {
