@@ -42,6 +42,9 @@
 
 #include "gdm-common.h"
 
+#include "gdm-settings-client.h"
+#include "gdm-settings-keys.h"
+
 #include "gdm-simple-slave.h"
 #include "gdm-simple-slave-glue.h"
 
@@ -60,7 +63,6 @@
 
 struct GdmSimpleSlavePrivate
 {
-        char              *id;
         GPid               pid;
 
         guint              greeter_reset_id;
@@ -75,6 +77,9 @@ struct GdmSimpleSlavePrivate
         GdmGreeterServer  *greeter_server;
         GdmGreeterSession *greeter;
         GdmSessionDirect  *session;
+
+        guint              start_session_when_ready : 1;
+        guint              waiting_to_start_session : 1;
 };
 
 enum {
@@ -87,11 +92,14 @@ static void     gdm_simple_slave_finalize       (GObject             *object);
 
 G_DEFINE_TYPE (GdmSimpleSlave, gdm_simple_slave, GDM_TYPE_SLAVE)
 
+static void create_new_session (GdmSimpleSlave *slave);
+
 static void
 on_session_started (GdmSession       *session,
+                    int               pid,
                     GdmSimpleSlave   *slave)
 {
-        g_debug ("GdmSimpleSlave: session started");
+        g_debug ("GdmSimpleSlave: session started %d", pid);
 
         /* FIXME: should we do something here? */
 }
@@ -134,10 +142,30 @@ add_user_authorization (GdmSimpleSlave *slave,
         return ret;
 }
 
+static void
+destroy_session (GdmSimpleSlave *slave)
+{
+        if (slave->priv->session != NULL) {
+                gdm_session_close (GDM_SESSION (slave->priv->session));
+                g_object_unref (slave->priv->session);
+                slave->priv->session = NULL;
+        }
+}
+
+static void
+reset_session (GdmSimpleSlave *slave)
+{
+        destroy_session (slave);
+        create_new_session (slave);
+        gdm_session_open (GDM_SESSION (slave->priv->session));
+}
+
 static gboolean
 greeter_reset_timeout (GdmSimpleSlave *slave)
 {
+        g_debug ("GdmSimpleSlave: resetting greeter");
         gdm_greeter_server_reset (slave->priv->greeter_server);
+        reset_session (slave);
         slave->priv->greeter_reset_id = 0;
         return FALSE;
 }
@@ -166,6 +194,7 @@ on_session_setup_failed (GdmSession     *session,
 {
         gdm_greeter_server_problem (slave->priv->greeter_server, _("Unable to initialize login system"));
 
+        destroy_session (slave);
         queue_greeter_reset (slave);
 }
 
@@ -197,19 +226,44 @@ on_session_authentication_failed (GdmSession     *session,
                                   GdmSimpleSlave *slave)
 {
         gdm_greeter_server_problem (slave->priv->greeter_server, _("Unable to authenticate user"));
-
+        destroy_session (slave);
         queue_greeter_reset (slave);
 }
 
+static void
+gdm_simple_slave_accredit_when_ready (GdmSimpleSlave *slave)
+{
+        if (slave->priv->start_session_when_ready) {
+                char *ssid;
+                char *username;
+                int   cred_flag;
+
+                slave->priv->waiting_to_start_session = FALSE;
+
+                username = gdm_session_direct_get_username (slave->priv->session);
+
+                ssid = gdm_slave_get_primary_session_id_for_user (GDM_SLAVE (slave), username);
+                if (ssid != NULL && ssid [0] != '\0') {
+                        /* FIXME: we don't yet support refresh */
+                        cred_flag = GDM_SESSION_CRED_ESTABLISH;
+                } else {
+                        cred_flag = GDM_SESSION_CRED_ESTABLISH;
+                }
+                g_free (ssid);
+                g_free (username);
+
+                gdm_session_accredit (GDM_SESSION (slave->priv->session), cred_flag);
+        } else {
+                slave->priv->waiting_to_start_session = TRUE;
+        }
+}
+
 static void
 on_session_authorized (GdmSession     *session,
                        GdmSimpleSlave *slave)
 {
-        int flag;
-
-        flag = GDM_SESSION_CRED_ESTABLISH;
-
-        gdm_session_accredit (session, flag);
+        gdm_greeter_server_user_authorized (slave->priv->greeter_server);
+        gdm_simple_slave_accredit_when_ready (slave);
 }
 
 static void
@@ -219,6 +273,7 @@ on_session_authorization_failed (GdmSession     *session,
 {
         gdm_greeter_server_problem (slave->priv->greeter_server, _("Unable to authorize user"));
 
+        destroy_session (slave);
         queue_greeter_reset (slave);
 }
 
@@ -252,7 +307,14 @@ start_session_timeout (GdmSimpleSlave *slave)
         migrated = try_migrate_session (slave);
         g_debug ("GdmSimpleSlave: migrated: %d", migrated);
         if (migrated) {
+                destroy_session (slave);
+
+                /* We don't stop the slave here because
+                   when Xorg exits it switches to the VT it was
+                   started from.  That interferes with fast
+                   user switching. */
                 queue_greeter_reset (slave);
+
                 goto out;
         }
 
@@ -310,6 +372,12 @@ on_session_accreditation_failed (GdmSession     *session,
                 gdm_greeter_server_problem (slave->priv->greeter_server, _("Unable establish credentials"));
         }
 
+        /* We don't stop the slave here after migrating because
+           when Xorg exits it switches to the VT it was
+           started from.  That interferes with fast
+           user switching. */
+        destroy_session (slave);
+
         queue_greeter_reset (slave);
 }
 
@@ -350,17 +418,50 @@ on_session_secret_info_query (GdmSession     *session,
         gdm_greeter_server_secret_info_query (slave->priv->greeter_server, text);
 }
 
+static gboolean
+get_timed_login_details (GdmSimpleSlave  *slave,
+                         char           **username,
+                         int             *delay)
+{
+        gboolean enabled;
+        gboolean res;
+
+        enabled = FALSE;
+        res = gdm_settings_client_get_boolean (GDM_KEY_TIMED_LOGIN_ENABLE, &enabled);
+        if (! enabled) {
+                return FALSE;
+        }
+
+        *username = NULL;
+        res = gdm_settings_client_get_string (GDM_KEY_TIMED_LOGIN_USER, username);
+
+        if (username == NULL) {
+                return FALSE;
+        }
+
+        res = gdm_settings_client_get_int (GDM_KEY_TIMED_LOGIN_DELAY, delay);
+
+        return TRUE;
+}
+
 static void
 on_session_opened (GdmSession     *session,
                    GdmSimpleSlave *slave)
 {
         gboolean res;
+        char    *username;
+        int      delay;
 
         g_debug ("GdmSimpleSlave: session opened");
         res = gdm_greeter_server_ready (slave->priv->greeter_server);
         if (! res) {
                 g_warning ("Unable to send ready");
         }
+
+        if (get_timed_login_details (slave, &username, &delay)) {
+                gdm_greeter_server_request_timed_login (slave->priv->greeter_server, username, delay);
+                g_free (username);
+        }
 }
 
 static void
@@ -373,11 +474,31 @@ on_session_selected_user_changed (GdmSession     *session,
         gdm_greeter_server_selected_user_changed (slave->priv->greeter_server, text);
 }
 
+static void
+on_default_language_name_changed (GdmSession     *session,
+                                  const char     *text,
+                                  GdmSimpleSlave *slave)
+{
+        g_debug ("GdmSimpleSlave: Default language name changed: %s", text);
+
+        gdm_greeter_server_default_language_name_changed (slave->priv->greeter_server, text);
+}
+
+static void
+on_default_session_name_changed (GdmSession     *session,
+                                 const char     *text,
+                                 GdmSimpleSlave *slave)
+{
+        g_debug ("GdmSimpleSlave: Default session name changed: %s", text);
+
+        gdm_greeter_server_default_session_name_changed (slave->priv->greeter_server, text);
+}
 
 static void
 create_new_session (GdmSimpleSlave *slave)
 {
         gboolean       display_is_local;
+        char          *display_id;
         char          *display_name;
         char          *display_hostname;
         char          *display_device;
@@ -386,6 +507,7 @@ create_new_session (GdmSimpleSlave *slave)
         g_debug ("GdmSimpleSlave: Creating new session");
 
         g_object_get (slave,
+                      "display-id", &display_id,
                       "display-name", &display_name,
                       "display-hostname", &display_hostname,
                       "display-is-local", &display_is_local,
@@ -397,11 +519,13 @@ create_new_session (GdmSimpleSlave *slave)
                 display_device = gdm_server_get_display_device (slave->priv->server);
         }
 
-        slave->priv->session = gdm_session_direct_new (display_name,
+        slave->priv->session = gdm_session_direct_new (display_id,
+                                                       display_name,
                                                        display_hostname,
                                                        display_device,
                                                        display_x11_authority_file,
                                                        display_is_local);
+        g_free (display_id);
         g_free (display_name);
         g_free (display_device);
         g_free (display_hostname);
@@ -488,6 +612,16 @@ create_new_session (GdmSimpleSlave *slave)
                           "selected-user-changed",
                           G_CALLBACK (on_session_selected_user_changed),
                           slave);
+
+        g_signal_connect (slave->priv->session,
+                          "default-language-name-changed",
+                          G_CALLBACK (on_default_language_name_changed),
+                          slave);
+
+        g_signal_connect (slave->priv->session,
+                          "default-session-name-changed",
+                          G_CALLBACK (on_default_session_name_changed),
+                          slave);
 }
 
 static void
@@ -528,7 +662,19 @@ on_greeter_begin_verification (GdmGreeterServer *greeter_server,
                                GdmSimpleSlave   *slave)
 {
         g_debug ("GdmSimpleSlave: begin verification");
-        gdm_session_setup (GDM_SESSION (slave->priv->session));
+        gdm_session_setup (GDM_SESSION (slave->priv->session),
+                           "gdm");
+}
+
+static void
+on_greeter_begin_auto_login (GdmGreeterServer *greeter_server,
+                             const char       *username,
+                             GdmSimpleSlave   *slave)
+{
+        g_debug ("GdmSimpleSlave: begin auto login for user '%s'", username);
+        gdm_session_setup_for_user (GDM_SESSION (slave->priv->session),
+                                    "gdm-autologin",
+                                    username);
 }
 
 static void
@@ -538,6 +684,7 @@ on_greeter_begin_verification_for_user (GdmGreeterServer *greeter_server,
 {
         g_debug ("GdmSimpleSlave: begin verification");
         gdm_session_setup_for_user (GDM_SESSION (slave->priv->session),
+                                    "gdm",
                                     username);
 }
 
@@ -578,15 +725,7 @@ on_greeter_cancel (GdmGreeterServer *greeter_server,
                    GdmSimpleSlave   *slave)
 {
         g_debug ("GdmSimpleSlave: Greeter cancelled");
-
-        if (slave->priv->session != NULL) {
-                gdm_session_close (GDM_SESSION (slave->priv->session));
-                g_object_unref (slave->priv->session);
-        }
-
-        create_new_session (slave);
-
-        gdm_session_open (GDM_SESSION (slave->priv->session));
+        reset_session (slave);
 }
 
 static void
@@ -595,7 +734,7 @@ on_greeter_connected (GdmGreeterServer *greeter_server,
 {
         gboolean display_is_local;
 
-        g_debug ("GdmSimpleSlave: Greeter started");
+        g_debug ("GdmSimpleSlave: Greeter connected");
 
         gdm_session_open (GDM_SESSION (slave->priv->session));
 
@@ -609,6 +748,26 @@ on_greeter_connected (GdmGreeterServer *greeter_server,
         }
 }
 
+static void
+on_start_session_when_ready (GdmGreeterServer *session,
+                             GdmSimpleSlave   *slave)
+{
+        g_debug ("GdmSimpleSlave: Will start session when ready");
+        slave->priv->start_session_when_ready = TRUE;
+
+        if (slave->priv->waiting_to_start_session) {
+                gdm_simple_slave_accredit_when_ready (slave);
+        }
+}
+
+static void
+on_start_session_later (GdmGreeterServer *session,
+                        GdmSimpleSlave   *slave)
+{
+        g_debug ("GdmSimpleSlave: Will start session when ready and told");
+        slave->priv->start_session_when_ready = FALSE;
+}
+
 static void
 setup_server (GdmSimpleSlave *slave)
 {
@@ -658,11 +817,15 @@ run_greeter (GdmSimpleSlave *slave)
         }
 
         /* Run the init script. gdmslave suspends until script has terminated */
-        gdm_slave_run_script (GDM_SLAVE (slave), GDMCONFDIR "/Init", "gdm");
+        gdm_slave_run_script (GDM_SLAVE (slave), GDMCONFDIR "/Init/Default", "gdm");
 
         create_new_session (slave);
 
         slave->priv->greeter_server = gdm_greeter_server_new (display_id);
+        g_signal_connect (slave->priv->greeter_server,
+                          "begin-auto-login",
+                          G_CALLBACK (on_greeter_begin_auto_login),
+                          slave);
         g_signal_connect (slave->priv->greeter_server,
                           "begin-verification",
                           G_CALLBACK (on_greeter_begin_verification),
@@ -695,6 +858,16 @@ run_greeter (GdmSimpleSlave *slave)
                           "cancelled",
                           G_CALLBACK (on_greeter_cancel),
                           slave);
+        g_signal_connect (slave->priv->greeter_server,
+                          "start-session-when-ready",
+                          G_CALLBACK (on_start_session_when_ready),
+                          slave);
+
+        g_signal_connect (slave->priv->greeter_server,
+                          "start-session-later",
+                          G_CALLBACK (on_start_session_later),
+                          slave);
+
         gdm_greeter_server_start (slave->priv->greeter_server);
 
         address = gdm_greeter_server_get_address (slave->priv->greeter_server);
@@ -845,9 +1018,7 @@ gdm_simple_slave_run (GdmSimpleSlave *slave)
 static gboolean
 gdm_simple_slave_start (GdmSlave *slave)
 {
-        gboolean res;
-
-        res = GDM_SLAVE_CLASS (gdm_simple_slave_parent_class)->start (slave);
+        GDM_SLAVE_CLASS (gdm_simple_slave_parent_class)->start (slave);
 
         gdm_simple_slave_run (GDM_SIMPLE_SLAVE (slave));
 
@@ -857,11 +1028,9 @@ gdm_simple_slave_start (GdmSlave *slave)
 static gboolean
 gdm_simple_slave_stop (GdmSlave *slave)
 {
-        gboolean res;
-
         g_debug ("GdmSimpleSlave: Stopping simple_slave");
 
-        res = GDM_SLAVE_CLASS (gdm_simple_slave_parent_class)->stop (slave);
+        GDM_SLAVE_CLASS (gdm_simple_slave_parent_class)->stop (slave);
 
         if (GDM_SIMPLE_SLAVE (slave)->priv->greeter != NULL) {
                 gdm_welcome_session_stop (GDM_WELCOME_SESSION (GDM_SIMPLE_SLAVE (slave)->priv->greeter));
@@ -890,10 +1059,6 @@ gdm_simple_slave_set_property (GObject      *object,
                                const GValue *value,
                                GParamSpec   *pspec)
 {
-        GdmSimpleSlave *self;
-
-        self = GDM_SIMPLE_SLAVE (object);
-
         switch (prop_id) {
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -907,10 +1072,6 @@ gdm_simple_slave_get_property (GObject    *object,
                                GValue      *value,
                                GParamSpec *pspec)
 {
-        GdmSimpleSlave *self;
-
-        self = GDM_SIMPLE_SLAVE (object);
-
         switch (prop_id) {
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -924,9 +1085,6 @@ gdm_simple_slave_constructor (GType                  type,
                               GObjectConstructParam *construct_properties)
 {
         GdmSimpleSlave      *simple_slave;
-        GdmSimpleSlaveClass *klass;
-
-        klass = GDM_SIMPLE_SLAVE_CLASS (g_type_class_peek (GDM_TYPE_SIMPLE_SLAVE));
 
         simple_slave = GDM_SIMPLE_SLAVE (G_OBJECT_CLASS (gdm_simple_slave_parent_class)->constructor (type,
                                                                                  n_construct_properties,
