@@ -34,6 +34,8 @@
 #include <dbus/dbus-glib-lowlevel.h>
 
 #include "gdm-greeter-client.h"
+#include "gdm-marshal.h"
+#include "gdm-profile.h"
 
 #define GDM_GREETER_CLIENT_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GDM_TYPE_GREETER_CLIENT, GdmGreeterClientPrivate))
 
@@ -66,6 +68,8 @@ enum {
         SELECTED_USER_CHANGED,
         DEFAULT_LANGUAGE_NAME_CHANGED,
         DEFAULT_SESSION_NAME_CHANGED,
+        TIMED_LOGIN_REQUESTED,
+        USER_AUTHORIZED,
         LAST_SIGNAL
 };
 
@@ -98,6 +102,36 @@ gdm_greeter_client_get_display_is_local (GdmGreeterClient *client)
         return client->priv->display_is_local;
 }
 
+static void
+emit_string_and_int_signal_for_message (GdmGreeterClient *client,
+                                        const char       *name,
+                                        DBusMessage      *message,
+                                        int               signal)
+{
+        DBusError   error;
+        const char *text;
+        int         number;
+        dbus_bool_t res;
+
+        dbus_error_init (&error);
+        res = dbus_message_get_args (message,
+                                     &error,
+                                     DBUS_TYPE_STRING, &text,
+                                     DBUS_TYPE_INT32, &number,
+                                     DBUS_TYPE_INVALID);
+        if (res) {
+
+                g_debug ("GdmGreeterClient: Received %s (%s %d)", name, text, number);
+
+                g_signal_emit (client,
+                               gdm_greeter_client_signals[signal],
+                               0, text, number);
+        } else {
+                g_warning ("Unable to get arguments: %s", error.message);
+                dbus_error_free (&error);
+        }
+}
+
 static void
 emit_string_signal_for_message (GdmGreeterClient *client,
                                 const char       *name,
@@ -147,6 +181,22 @@ on_default_session_name_changed (GdmGreeterClient *client,
         emit_string_signal_for_message (client, "DefaultSessionNameChanged", message, DEFAULT_SESSION_NAME_CHANGED);
 }
 
+static void
+on_timed_login_requested (GdmGreeterClient *client,
+                          DBusMessage      *message)
+{
+        emit_string_and_int_signal_for_message (client, "TimedLoginRequested", message, TIMED_LOGIN_REQUESTED);
+}
+
+static void
+on_user_authorized (GdmGreeterClient *client,
+                    DBusMessage      *message)
+{
+        g_signal_emit (client,
+                       gdm_greeter_client_signals[USER_AUTHORIZED],
+                       0);
+}
+
 static void
 on_info_query (GdmGreeterClient *client,
                DBusMessage      *message)
@@ -244,12 +294,61 @@ send_dbus_string_method (DBusConnection *connection,
                            error.message);
                 return FALSE;
         }
-        dbus_message_unref (reply);
+        if (reply != NULL) {
+                dbus_message_unref (reply);
+        }
         dbus_connection_flush (connection);
 
         return TRUE;
 }
 
+static gboolean
+send_dbus_bool_method (DBusConnection *connection,
+                       const char     *method,
+                       gboolean        payload)
+{
+        DBusError       error;
+        DBusMessage    *message;
+        DBusMessage    *reply;
+        DBusMessageIter iter;
+
+        g_debug ("GdmGreeterClient: Calling %s", method);
+        message = dbus_message_new_method_call (NULL,
+                                                GREETER_SERVER_DBUS_PATH,
+                                                GREETER_SERVER_DBUS_INTERFACE,
+                                                method);
+        if (message == NULL) {
+                g_warning ("Couldn't allocate the D-Bus message");
+                return FALSE;
+        }
+
+        dbus_message_iter_init_append (message, &iter);
+        dbus_message_iter_append_basic (&iter,
+                                        DBUS_TYPE_BOOLEAN,
+                                        &payload);
+
+        dbus_error_init (&error);
+        reply = dbus_connection_send_with_reply_and_block (connection,
+                                                           message,
+                                                           -1,
+                                                           &error);
+
+        dbus_message_unref (message);
+
+        if (dbus_error_is_set (&error)) {
+                g_warning ("%s %s raised: %s\n",
+                           method,
+                           error.name,
+                           error.message);
+                return FALSE;
+        }
+        if (reply != NULL) {
+                dbus_message_unref (reply);
+        }
+        dbus_connection_flush (connection);
+
+        return TRUE;
+}
 
 static gboolean
 send_dbus_void_method (DBusConnection *connection,
@@ -284,17 +383,20 @@ send_dbus_void_method (DBusConnection *connection,
                            error.message);
                 return FALSE;
         }
-        dbus_message_unref (reply);
+        if (reply != NULL) {
+                dbus_message_unref (reply);
+        }
         dbus_connection_flush (connection);
 
         return TRUE;
 }
 
 void
-gdm_greeter_client_call_begin_timed_login (GdmGreeterClient *client)
+gdm_greeter_client_call_begin_auto_login (GdmGreeterClient *client,
+                                          const char       *username)
 {
-        send_dbus_void_method (client->priv->connection,
-                               "BeginTimedLogin");
+        send_dbus_string_method (client->priv->connection,
+                                 "BeginAutoLogin", username);
 }
 
 void
@@ -322,6 +424,15 @@ gdm_greeter_client_call_answer_query (GdmGreeterClient *client,
                                  text);
 }
 
+void
+gdm_greeter_client_call_start_session_when_ready  (GdmGreeterClient *client,
+                                                   gboolean          should_start_session)
+{
+        send_dbus_bool_method (client->priv->connection,
+                               "StartSessionWhenReady",
+                               should_start_session);
+}
+
 void
 gdm_greeter_client_call_select_session (GdmGreeterClient *client,
                                         const char       *text)
@@ -524,6 +635,10 @@ client_dbus_handle_message (DBusConnection *connection,
                 on_default_language_name_changed (client, message);
         } else if (dbus_message_is_signal (message, GREETER_SERVER_DBUS_INTERFACE, "DefaultSessionNameChanged")) {
                 on_default_session_name_changed (client, message);
+        } else if (dbus_message_is_signal (message, GREETER_SERVER_DBUS_INTERFACE, "TimedLoginRequested")) {
+                on_timed_login_requested (client, message);
+        } else if (dbus_message_is_signal (message, GREETER_SERVER_DBUS_INTERFACE, "UserAuthorized")) {
+                on_user_authorized (client, message);
         } else {
                 return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
         }
@@ -574,6 +689,8 @@ gdm_greeter_client_start (GdmGreeterClient *client,
 
         g_return_val_if_fail (GDM_IS_GREETER_CLIENT (client), FALSE);
 
+        gdm_profile_start (NULL);
+
         ret = FALSE;
 
         if (client->priv->address == NULL) {
@@ -616,6 +733,8 @@ gdm_greeter_client_start (GdmGreeterClient *client,
         ret = TRUE;
 
  out:
+        gdm_profile_end (NULL);
+
         return ret;
 }
 
@@ -786,6 +905,27 @@ gdm_greeter_client_class_init (GdmGreeterClientClass *klass)
                               g_cclosure_marshal_VOID__STRING,
                               G_TYPE_NONE,
                               1, G_TYPE_STRING);
+
+        gdm_greeter_client_signals[TIMED_LOGIN_REQUESTED] =
+                g_signal_new ("timed-login-requested",
+                              G_OBJECT_CLASS_TYPE (object_class),
+                              G_SIGNAL_RUN_FIRST,
+                              G_STRUCT_OFFSET (GdmGreeterClientClass, default_session_name_changed),
+                              NULL,
+                              NULL,
+                              gdm_marshal_VOID__STRING_INT,
+                              G_TYPE_NONE,
+                              2, G_TYPE_STRING, G_TYPE_INT);
+
+        gdm_greeter_client_signals[USER_AUTHORIZED] =
+                g_signal_new ("user-authorized",
+                              G_OBJECT_CLASS_TYPE (object_class),
+                              G_SIGNAL_RUN_FIRST,
+                              G_STRUCT_OFFSET (GdmGreeterClientClass, user_authorized),
+                              NULL,
+                              NULL,
+                              g_cclosure_marshal_VOID__VOID,
+                              G_TYPE_NONE, 0);
 }
 
 static void
