@@ -30,6 +30,7 @@
 #include <errno.h>
 #include <dirent.h>
 #include <locale.h>
+#include <langinfo.h>
 #include <sys/stat.h>
 
 #include <glib.h>
@@ -50,36 +51,58 @@
 #define ISO_CODES_LOCALESDIR ISO_CODES_PREFIX "/share/locale"
 
 typedef struct _GdmLocale {
+        char *id;
         char *name;
         char *language_code;
         char *territory_code;
         char *codeset;
         char *modifier;
-        char *title;
-        char *language;
-        char *territory;
 } GdmLocale;
 
 static GHashTable *gdm_languages_map;
 static GHashTable *gdm_territories_map;
 static GHashTable *gdm_available_locales_map;
 
-static gboolean gdm_languages_is_initialized = FALSE;
-
 static void
-chooser_locale_free (GdmLocale *locale)
+gdm_locale_free (GdmLocale *locale)
 {
         if (locale == NULL) {
                 return;
         }
 
+        g_free (locale->id);
         g_free (locale->name);
-        g_free (locale->title);
-        g_free (locale->language);
-        g_free (locale->territory);
+        g_free (locale->codeset);
+        g_free (locale->modifier);
         g_free (locale);
 }
 
+static char *
+normalize_codeset (const char *codeset)
+{
+        char *normalized_codeset;
+        const char *p;
+        char *q;
+
+        normalized_codeset = g_strdup (codeset);
+
+        if (codeset != NULL) {
+                for (p = codeset, q = normalized_codeset;
+                     *p != '\0'; p++) {
+
+                        if (*p == '-' || *p == '_') {
+                                continue;
+                        }
+
+                        *q = g_ascii_tolower (*p);
+                        q++;
+                }
+                *q = '\0';
+        }
+
+        return normalized_codeset;
+}
+
 /*
  * According to http://en.wikipedia.org/wiki/Locale
  * locale names are of the form:
@@ -98,12 +121,23 @@ gdm_parse_language_name (const char *name,
         GError     *error;
 
         error = NULL;
-        re = g_regex_new ("(?P<language>[a-zA-Z]+)(_(?P<territory>[a-zA-Z]+))?(.(?P<codeset>[0-9a-zA-Z]+))?(@(?P<modifier>[0-9a-zA-Z]+))?", 0, 0, &error);
+        re = g_regex_new ("^(?P<language>[^_.@[:space:]]+)"
+                          "(_(?P<territory>[[:upper:]]+))?"
+                          "(.(?P<codeset>[-_0-9a-zA-Z]+))?"
+                          "(@(?P<modifier>[[:ascii:]]+))?$",
+                          0, 0, &error);
         if (re == NULL) {
                 g_critical (error->message);
+                return;
         }
 
-        g_regex_match (re, name, 0, &match_info);
+        if (!g_regex_match (re, name, 0, &match_info) ||
+            g_match_info_is_partial_match (match_info)) {
+                g_match_info_free (match_info);
+                g_regex_unref (re);
+                g_warning ("locale %s isn't valid\n", name);
+                return;
+        }
 
         res = g_match_info_matches (match_info);
         if (! res) {
@@ -117,14 +151,40 @@ gdm_parse_language_name (const char *name,
 
         if (territory_codep != NULL) {
                 *territory_codep = g_match_info_fetch_named (match_info, "territory");
+
+                if (*territory_codep != NULL &&
+                    *territory_codep[0] == '\0') {
+                        g_free (*territory_codep);
+                        *territory_codep = NULL;
+                }
         }
 
         if (codesetp != NULL) {
                 *codesetp = g_match_info_fetch_named (match_info, "codeset");
+
+                if (*codesetp != NULL &&
+                    *codesetp[0] == '\0') {
+                        g_free (*codesetp);
+                        *codesetp = NULL;
+                }
+
+                if (*codesetp != NULL) {
+                        char *codeset;
+
+                        codeset = normalize_codeset (*codesetp);
+                        g_free (*codesetp);
+                        *codesetp = codeset;
+                }
         }
 
         if (modifierp != NULL) {
                 *modifierp = g_match_info_fetch_named (match_info, "modifier");
+
+                if (*modifierp != NULL &&
+                    *modifierp[0] == '\0') {
+                        g_free (*modifierp);
+                        *modifierp = NULL;
+                }
         }
 
         g_match_info_free (match_info);
@@ -139,17 +199,19 @@ construct_language_name (const char *language,
 {
         char *name;
 
-        /* Ignore codeset and modifier for this */
-        /* FIXME: that can't be right. It's going to break
-         * serbian among other things.
-         */
-        if (territory == NULL) {
-                name = g_strdup (language);
-        } else {
-                name = g_strdup_printf ("%s_%s",
-                                        language,
-                                        territory);
-        }
+        g_assert (language[0] != 0);
+        g_assert (territory == NULL || territory[0] != 0);
+        g_assert (codeset == NULL || codeset[0] != 0);
+        g_assert (modifier == NULL || modifier[0] != 0);
+
+        name = g_strdup_printf ("%s%s%s%s%s%s%s",
+                                language,
+                                territory != NULL? "_" : "",
+                                territory != NULL? territory : "",
+                                codeset != NULL? "." : "",
+                                codeset != NULL? codeset : "",
+                                modifier != NULL? "@" : "",
+                                modifier != NULL? modifier : "");
 
         return name;
 }
@@ -160,80 +222,160 @@ gdm_normalize_language_name (const char *name)
         char *normalized_name;
         char *language_code;
         char *territory_code;
+        char *codeset;
+        char *modifier;
+
+        if (name[0] == '\0') {
+                return NULL;
+        }
 
         gdm_parse_language_name (name,
                                  &language_code,
                                  &territory_code,
-                                 NULL, NULL);
+                                 &codeset, &modifier);
 
         normalized_name = construct_language_name (language_code,
-                                                   territory_code, NULL, NULL);
+                                                   territory_code,
+                                                   codeset, modifier);
         g_free (language_code);
         g_free (territory_code);
+        g_free (codeset);
+        g_free (modifier);
 
         return normalized_name;
 }
 
-static char *
-utf8_convert (const char *str,
-              int         len)
+static gboolean
+language_name_is_valid (const char *language_name)
 {
-        char *utf8;
+        char     *old_locale;
+        gboolean  is_valid;
 
-        utf8 = g_locale_to_utf8 (str, len, NULL, NULL, NULL);
+        old_locale = g_strdup (setlocale (LC_MESSAGES, NULL));
+        is_valid = setlocale (LC_MESSAGES, language_name) != NULL;
+        setlocale (LC_MESSAGES, old_locale);
+        g_free (old_locale);
 
-        /* if we couldn't convert text from locale then
-         * assume utf-8 and hope for the best */
-        if (utf8 == NULL) {
-                char *p;
-                char *q;
-
-                if (len < 0) {
-                        utf8 = g_strdup (str);
-                } else {
-                        utf8 = g_strndup (str, len);
-                }
-
-                p = utf8;
-                while (*p != '\0' && !g_utf8_validate ((const char *)p, -1, (const char **)&q)) {
-                        *q = '?';
-                        p = q + 1;
-                }
-        }
-
-        return utf8;
+        return is_valid;
 }
 
-/* Magic number at the beginning of a locale data file for CATEGORY.  */
-#define LIMAGIC(category) \
-  (category == LC_COLLATE                                               \
-   ? ((unsigned int) (0x20051014 ^ (category)))                         \
-   : ((unsigned int) (0x20031115 ^ (category))))
-
-
-/* This seems to be specified by ISO/IEC 14652 */
-static void
-get_lc_identification (GdmLocale *locale,
-                       void             *data,
-                       gsize             size)
+static gboolean
+language_name_is_utf8 (const char *language_name)
 {
-        struct {
-                unsigned int magic;
-                unsigned int nstrings;
-                unsigned int strindex[];
-        } *filedata = data;
+        char     *old_locale;
+        char     *codeset;
+        gboolean  is_utf8;
 
-#ifdef LC_IDENTIFICATION
-        if (filedata->magic == LIMAGIC (LC_IDENTIFICATION)
-            && (sizeof *filedata + (filedata->nstrings * sizeof (unsigned int)) <= size)) {
+        old_locale = g_strdup (setlocale (LC_CTYPE, NULL));
 
-#define GET_HANDLE(idx) ((char *) data + filedata->strindex[_NL_ITEM_INDEX (_NL_IDENTIFICATION_##idx)])
-
-                locale->title = utf8_convert (GET_HANDLE (TITLE), -1);
-                locale->language = utf8_convert (GET_HANDLE (LANGUAGE), -1);
-                locale->territory = utf8_convert (GET_HANDLE (TERRITORY), -1);
+        if (setlocale (LC_CTYPE, language_name) == NULL) {
+                g_free (old_locale);
+                return FALSE;
         }
-#endif
+
+        codeset = normalize_codeset (nl_langinfo (CODESET));
+
+        is_utf8 = strcmp (codeset, "utf8") == 0;
+        g_free (codeset);
+
+        setlocale (LC_CTYPE, old_locale);
+        g_free (old_locale);
+
+        return is_utf8;
+}
+
+static gboolean
+language_name_has_translations (const char *language_name)
+{
+        GDir        *dir;
+        char        *path;
+        const char  *name;
+        gboolean     has_translations;
+
+        path = g_build_filename (GNOMELOCALEDIR, language_name, "LC_MESSAGES", NULL);
+
+        has_translations = FALSE;
+        dir = g_dir_open (path, 0, NULL);
+        g_free (path);
+
+        if (dir == NULL) {
+                goto out;
+        }
+
+        do {
+                name = g_dir_read_name (dir);
+
+                if (name == NULL) {
+                        break;
+                }
+
+                if (g_str_has_suffix (name, ".mo")) {
+                        has_translations = TRUE;
+                        break;
+                }
+        } while (name != NULL);
+        g_dir_close (dir);
+
+out:
+        return has_translations;
+}
+
+static gboolean
+add_locale (const char *language_name)
+{
+        GdmLocale *locale;
+        GdmLocale *old_locale;
+        char      *name;
+
+        if (language_name_is_utf8 (language_name)) {
+                name = g_strdup (language_name);
+        } else {
+                name = g_strdup_printf ("%s.utf8", language_name);
+
+                if (!language_name_is_utf8 (name)) {
+                        g_free (name);
+                        return FALSE;
+                }
+        }
+
+        if (!language_name_is_valid (name)) {
+                g_free (name);
+                return FALSE;
+        }
+
+
+        locale = g_new0 (GdmLocale, 1);
+        gdm_parse_language_name (name,
+                                 &locale->language_code,
+                                 &locale->territory_code,
+                                 &locale->codeset,
+                                 &locale->modifier);
+        g_free (name);
+        name = NULL;
+
+        locale->id = construct_language_name (locale->language_code, locale->territory_code,
+                                              NULL, locale->modifier);
+        locale->name = construct_language_name (locale->language_code, locale->territory_code,
+                                                locale->codeset, locale->modifier);
+
+        if (!language_name_has_translations (locale->name) &&
+            !language_name_has_translations (locale->id) &&
+            !language_name_has_translations (locale->language_code)) {
+                gdm_locale_free (locale);
+                return FALSE;
+        }
+
+        old_locale = g_hash_table_lookup (gdm_available_locales_map, locale->id);
+        if (old_locale != NULL) {
+                if (strlen (old_locale->name) > strlen (locale->name)) {
+                        gdm_locale_free (locale);
+                        return FALSE;
+                }
+        }
+
+        g_hash_table_insert (gdm_available_locales_map, g_strdup (locale->id), locale);
+
+        return TRUE;
 }
 
 struct nameent
@@ -242,14 +384,7 @@ struct nameent
         uint32_t locrec_offset;
 };
 
-static int
-nameentcmp (const void *a, const void *b)
-{
-        return strcoll (((const struct nameent *) a)->name,
-                        ((const struct nameent *) b)->name);
-}
-
-static void
+static gboolean
 collect_locales_from_archive (void)
 {
         GMappedFile        *mapped;
@@ -261,15 +396,18 @@ collect_locales_from_archive (void)
         uint32_t            used;
         uint32_t            cnt;
         gsize               len;
+        gboolean            locales_collected;
 
         error = NULL;
         mapped = g_mapped_file_new (ARCHIVE_FILE, FALSE, &error);
         if (mapped == NULL) {
                 g_warning ("Mapping failed for %s: %s", ARCHIVE_FILE, error->message);
                 g_error_free (error);
-                return;
+                return FALSE;
         }
 
+        locales_collected = FALSE;
+
         addr = g_mapped_file_get_contents (mapped);
         len = g_mapped_file_get_length (mapped);
 
@@ -291,43 +429,17 @@ collect_locales_from_archive (void)
                 }
         }
 
-        /* Sort the names.  */
-        qsort (names, used, sizeof (struct nameent), nameentcmp);
-
         for (cnt = 0; cnt < used; ++cnt) {
-                struct locrecent *locrec;
-                GdmLocale *locale;
-
-                locale = g_new0 (GdmLocale, 1);
-
-                gdm_parse_language_name (names[cnt].name,
-                                         &locale->language_code,
-                                         &locale->territory_code,
-                                         NULL, NULL);
-
-                locale->name = construct_language_name (locale->language_code, locale->territory_code, NULL, NULL);
-
-                if (g_hash_table_lookup (gdm_available_locales_map, locale->name) != NULL) {
-                        chooser_locale_free (locale);
-                        continue;
-                }
-
-                locrec = (struct locrecent *) (addr + names[cnt].locrec_offset);
-
-#ifdef LC_IDENTIFICATION
-                get_lc_identification (locale,
-                                       addr + locrec->record[LC_IDENTIFICATION].offset,
-                                       locrec->record[LC_IDENTIFICATION].len);
-#endif
-
-                g_hash_table_insert (gdm_available_locales_map, g_strdup (locale->name), locale);
+                add_locale (names[cnt].name);
         }
 
         g_free (names);
 
+        locales_collected = TRUE;
  out:
 
         g_mapped_file_free (mapped);
+        return locales_collected;
 }
 
 static int
@@ -347,7 +459,7 @@ select_dirs (const struct dirent *dirent)
                                 struct stat st;
                                 char       *path;
 
-                                path = g_build_filename (LIBLOCALEDIR, dirent->d_name, NULL);
+                                path = g_build_filename (GNOMELOCALEDIR, dirent->d_name, NULL);
                                 if (g_stat (path, &st) == 0) {
                                         mode = st.st_mode;
                                 }
@@ -367,49 +479,10 @@ collect_locales_from_directory (void)
         int             ndirents;
         int             cnt;
 
-        ndirents = scandir (LIBLOCALEDIR, &dirents, select_dirs, alphasort);
+        ndirents = scandir (GNOMELOCALEDIR, &dirents, select_dirs, alphasort);
 
         for (cnt = 0; cnt < ndirents; ++cnt) {
-                char             *path;
-                GdmLocale *locale;
-                gboolean          res;
-
-                locale = g_new0 (GdmLocale, 1);
-                gdm_parse_language_name (dirents[cnt]->d_name,
-                                         &locale->language_code,
-                                         &locale->territory_code,
-                                         NULL, NULL);
-
-                /* Ignore codeset and modifier for this */
-                locale->name = construct_language_name (locale->language_code, locale->territory_code, NULL, NULL);
-
-                if (g_hash_table_lookup (gdm_available_locales_map, locale->name) != NULL) {
-                        chooser_locale_free (locale);
-                        continue;
-                }
-
-                /* try to get additional information from LC_IDENTIFICATION */
-                path = g_build_filename (LIBLOCALEDIR, dirents[cnt]->d_name, "LC_IDENTIFICATION", NULL);
-                res = g_file_test (path, G_FILE_TEST_IS_REGULAR);
-                if (res) {
-                        GMappedFile      *mapped;
-                        GError           *error;
-
-                        error = NULL;
-                        mapped = g_mapped_file_new (path, FALSE, &error);
-                        if (mapped == NULL) {
-                                g_warning ("Mapping failed for %s: %s", path, error->message);
-                                g_error_free (error);
-                        } else {
-                                get_lc_identification (locale,
-                                                       g_mapped_file_get_contents (mapped),
-                                                       g_mapped_file_get_length (mapped));
-                                g_mapped_file_free (mapped);
-                        }
-                }
-                g_free (path);
-
-                g_hash_table_insert (gdm_available_locales_map, g_strdup (locale->name), locale);
+                add_locale (dirents[cnt]->d_name);
         }
 
         if (ndirents > 0) {
@@ -417,23 +490,38 @@ collect_locales_from_directory (void)
         }
 }
 
-static void
-collect_locales_from_aliases (void)
-{
-        /* FIXME: */
-}
-
 static void
 collect_locales (void)
 {
 
         if (gdm_available_locales_map == NULL) {
-                gdm_available_locales_map = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, (GDestroyNotify) chooser_locale_free);
+                gdm_available_locales_map = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, (GDestroyNotify) gdm_locale_free);
         }
 
-        collect_locales_from_archive ();
-        collect_locales_from_directory ();
-        collect_locales_from_aliases ();
+        if (collect_locales_from_archive ()) {
+                return;
+        } else {
+                g_warning ("Could not read list of available locales from libc, "
+                           "guessing possible locales from available translations, "
+                           "but list may be incomplete!");
+
+                collect_locales_from_directory ();
+        }
+}
+
+static gboolean
+is_fallback_language (const char *code)
+{
+        const char *fallback_language_names[] = { "C", "POSIX", NULL };
+        int i;
+
+        for (i = 0; fallback_language_names[i] != NULL; i++) {
+                if (strcmp (code, fallback_language_names[i]) == 0) {
+                        return TRUE;
+                }
+        }
+
+        return FALSE;
 }
 
 static const char *
@@ -444,6 +532,10 @@ get_language (const char *code)
 
         g_assert (code != NULL);
 
+        if (is_fallback_language (code)) {
+                return "Unspecified";
+        }
+
         len = strlen (code);
         if (len != 2 && len != 3) {
                 return NULL;
@@ -454,6 +546,24 @@ get_language (const char *code)
         return name;
 }
 
+static char *
+get_first_item_in_semicolon_list (const char *list)
+{
+    char **items;
+    char  *item;
+
+    /* Some entries in iso codes have multiple values, separated
+     * by semicolons.  Not really sure which one to pick, so
+     * we just arbitrarily pick the first one.
+     */
+    items = g_strsplit (list, "; ", 2);
+
+    item = g_strdup (items[0]);
+    g_strfreev (items);
+
+    return item;
+}
+
 static char *
 get_translated_language (const char *code,
                          const char *locale)
@@ -465,20 +575,25 @@ get_translated_language (const char *code,
 
         name = NULL;
         if (language != NULL) {
-                const char *translated_name;
-                char *old_locale;
-                char **names;
+                const char  *translated_name;
+                char        *old_locale;
 
-                old_locale = g_strdup (setlocale (LC_MESSAGES, NULL));
-                setlocale (LC_MESSAGES, locale);
-                translated_name = dgettext ("iso_639", language);
-                setlocale (LC_MESSAGES, old_locale);
-                g_free (old_locale);
+                if (locale != NULL) {
+                        old_locale = g_strdup (setlocale (LC_MESSAGES, NULL));
+                        setlocale (LC_MESSAGES, locale);
+                }
 
-                names = g_strsplit (translated_name, "; ", 2);
+                if (is_fallback_language (code)) {
+                        name = _("Unspecified");
+                } else {
+                        translated_name = dgettext ("iso_639", language);
+                        name = get_first_item_in_semicolon_list (translated_name);
+                }
 
-                name = g_strdup (names[0]);
-                g_strfreev (names);
+                if (locale != NULL) {
+                        setlocale (LC_MESSAGES, old_locale);
+                        g_free (old_locale);
+                }
         }
 
         return name;
@@ -514,19 +629,20 @@ get_translated_territory (const char *code,
         name = NULL;
         if (territory != NULL) {
                 const char *translated_territory;
-                char *old_locale;
-                char **territories;
+                char       *old_locale;
+
+                if (locale != NULL) {
+                        old_locale = g_strdup (setlocale (LC_MESSAGES, NULL));
+                        setlocale (LC_MESSAGES, locale);
+                }
 
-                old_locale = g_strdup (setlocale (LC_MESSAGES, NULL));
-                setlocale (LC_MESSAGES, locale);
                 translated_territory = dgettext ("iso_3166", territory);
-                setlocale (LC_MESSAGES, old_locale);
-                g_free (old_locale);
+                name = get_first_item_in_semicolon_list (translated_territory);
 
-                territories = g_strsplit (translated_territory, "; ", 2);
-
-                name = g_strdup (territories[0]);
-                g_strfreev (territories);
+                if (locale != NULL) {
+                        setlocale (LC_MESSAGES, old_locale);
+                        g_free (old_locale);
+                }
         }
 
         return name;
@@ -773,7 +889,8 @@ territories_init (void)
 }
 
 char *
-gdm_get_language_from_name (const char *name)
+gdm_get_language_from_name (const char *name,
+                            const char *locale)
 {
         char *full_language;
         char *language_code;
@@ -781,12 +898,12 @@ gdm_get_language_from_name (const char *name)
         const char *language;
         const char *territory;
 
-        if (!gdm_languages_is_initialized) {
-                collect_locales ();
+        if (gdm_languages_map == NULL) {
                 languages_init ();
-                territories_init ();
+        }
 
-                gdm_languages_is_initialized = TRUE;
+        if (gdm_territories_map == NULL) {
+                territories_init ();
         }
 
         language_code = NULL;
@@ -800,10 +917,10 @@ gdm_get_language_from_name (const char *name)
                 goto out;
         }
 
-        language = get_translated_language (language_code, name);
+        language = get_translated_language (language_code, locale);
 
         if (territory_code != NULL) {
-                territory = get_translated_territory (territory_code, name);
+                territory = get_translated_territory (territory_code, locale);
         } else {
                 territory = NULL;
         }
@@ -830,12 +947,8 @@ gdm_get_all_language_names (void)
         gpointer key, value;
         GPtrArray *array;
 
-        if (!gdm_languages_is_initialized) {
+        if (gdm_available_locales_map == NULL) {
                 collect_locales ();
-                languages_init ();
-                territories_init ();
-
-                gdm_languages_is_initialized = TRUE;
         }
 
         array = g_ptr_array_new ();
