@@ -68,6 +68,8 @@ struct GdmOptionWidgetPrivate
         gint                     number_of_top_rows;
         gint                     number_of_middle_rows;
         gint                     number_of_bottom_rows;
+
+        guint                    check_idle_id;
 };
 
 enum {
@@ -447,16 +449,31 @@ gdm_option_widget_dispose (GObject *object)
         G_OBJECT_CLASS (gdm_option_widget_parent_class)->dispose (object);
 }
 
+static gboolean
+gdm_option_widget_mnemonic_activate (GtkWidget *widget,
+                                     gboolean   group_cycling)
+{
+        GdmOptionWidget *option_widget;
+
+        option_widget = GDM_OPTION_WIDGET (widget);
+        gtk_widget_grab_focus (option_widget->priv->items_combo_box);
+        gtk_combo_box_popup (GTK_COMBO_BOX (option_widget->priv->items_combo_box));
+
+        return TRUE;
+}
+
 static void
 gdm_option_widget_class_init (GdmOptionWidgetClass *klass)
 {
         GObjectClass   *object_class = G_OBJECT_CLASS (klass);
+        GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
 
         object_class->get_property = gdm_option_widget_get_property;
         object_class->set_property = gdm_option_widget_set_property;
         object_class->constructor = gdm_option_widget_constructor;
         object_class->dispose = gdm_option_widget_dispose;
         object_class->finalize = gdm_option_widget_finalize;
+        widget_class->mnemonic_activate = gdm_option_widget_mnemonic_activate;
 
         gtk_rc_parse_string (GDM_OPTION_WIDGET_RC_STRING);
 
@@ -597,6 +614,39 @@ path_is_separator (GdmOptionWidget *widget,
                path_is_bottom_separator (widget, model, path);
 }
 
+static gboolean
+gdm_option_widget_check_visibility (GdmOptionWidget *widget)
+{
+        int number_of_rows;
+
+        number_of_rows = widget->priv->number_of_top_rows +
+                         widget->priv->number_of_middle_rows +
+                         widget->priv->number_of_bottom_rows;
+
+        if (number_of_rows > 1) {
+                gtk_widget_show (widget->priv->items_combo_box);
+                gtk_widget_show (widget->priv->label);
+
+                if (widget->priv->icon_name != NULL) {
+                        gtk_widget_show (widget->priv->image);
+                }
+        } else {
+                gtk_widget_hide (widget->priv->items_combo_box);
+                gtk_widget_hide (widget->priv->label);
+                gtk_widget_hide (widget->priv->image);
+        }
+
+        return FALSE;
+}
+
+static void
+gdm_option_widget_queue_visibility_check (GdmOptionWidget *widget)
+{
+        if (widget->priv->check_idle_id == 0) {
+                widget->priv->check_idle_id = g_idle_add ((GSourceFunc) gdm_option_widget_check_visibility, widget);
+        }
+}
+
 static gboolean
 check_item_visibilty (GtkTreeModel *model,
                       GtkTreeIter  *iter,
@@ -627,6 +677,8 @@ check_item_visibilty (GtkTreeModel *model,
                 is_visible = TRUE;
         }
 
+        gdm_option_widget_queue_visibility_check (widget);
+
         return is_visible;
 }
 
@@ -816,7 +868,6 @@ gdm_option_widget_init (GdmOptionWidget *widget)
         gtk_container_add (GTK_CONTAINER (widget),
                            box);
 
-
         widget->priv->image = gtk_image_new ();
         gtk_widget_set_no_show_all (widget->priv->image, TRUE);
         gtk_box_pack_start (GTK_BOX (box), widget->priv->image, FALSE, FALSE, 0);
@@ -824,7 +875,7 @@ gdm_option_widget_init (GdmOptionWidget *widget)
         widget->priv->label = gtk_label_new ("");
         gtk_label_set_use_underline (GTK_LABEL (widget->priv->label), TRUE);
         gtk_label_set_use_markup (GTK_LABEL (widget->priv->label), TRUE);
-        gtk_widget_show (widget->priv->label);
+        gtk_widget_set_no_show_all (widget->priv->label, TRUE);
         gtk_box_pack_start (GTK_BOX (box), widget->priv->label, FALSE, FALSE, 0);
 
         widget->priv->items_combo_box = gtk_combo_box_new ();
@@ -842,11 +893,11 @@ gdm_option_widget_init (GdmOptionWidget *widget)
                           G_CALLBACK (on_default_item_changed),
                           NULL);
 
-        gtk_widget_show (widget->priv->items_combo_box);
+        gtk_widget_set_no_show_all (widget->priv->items_combo_box, TRUE);
         gtk_container_add (GTK_CONTAINER (box),
                            widget->priv->items_combo_box);
         gtk_label_set_mnemonic_widget (GTK_LABEL (widget->priv->label),
-                                       widget->priv->items_combo_box);
+                                       GTK_WIDGET (widget));
 
         g_assert (NUMBER_OF_OPTION_COLUMNS == 4);
         widget->priv->list_store = gtk_list_store_new (NUMBER_OF_OPTION_COLUMNS,
@@ -1041,10 +1092,6 @@ gdm_option_widget_lookup_item (GdmOptionWidget          *widget,
         char         *active_item_id;
 
         g_return_val_if_fail (GDM_IS_OPTION_WIDGET (widget), FALSE);
-
-        if (id == NULL) {
-                sleep (30);
-        }
         g_return_val_if_fail (id != NULL, FALSE);
 
         active_item_id = get_active_item_id (widget, &iter);
